#pragma once
#pragma warning(default : 4710)


class HSVComparator {
public:
	static enum EVAL_ORDER {
		HSV, SVH, VHS, 
		VSH, HVS, SHV, 
	};


	static enum FUNC_TYPE {
		MIN, MAX, WARM1, COOL1, WARM2, COOL2, 
	};


	HSVComparator() {
		setEvalOrder(HSV);
		setEvalFunc_H(MIN);
		setEvalFunc_S(MIN);
		setEvalFunc_V(MIN);
		setHueBreakPoint(0);
	}

	virtual ~HSVComparator() {}

	
	inline HSVComparator::EVAL_ORDER getEvalOrder() { return eval_order; }
	inline HSVComparator::FUNC_TYPE getEvalFunc_H() { return h_func; }
	inline HSVComparator::FUNC_TYPE getEvalFunc_S() { return s_func; }
	inline HSVComparator::FUNC_TYPE getEvalFunc_V() { return v_func; }

	inline float getHueBreakPoint() { return h_breakpoint; }


	inline void setEvalOrder(HSVComparator::EVAL_ORDER order);
	inline void setEvalFunc_H(HSVComparator::FUNC_TYPE func);
	inline void setEvalFunc_S(HSVComparator::FUNC_TYPE func);
	inline void setEvalFunc_V(HSVComparator::FUNC_TYPE func);

	inline void setHueBreakPoint(float val);		// in WARM/COOL functions, it is used to specify a warm color: its complement will automatically be a cold color...


	// pass each single color represented in HSV color. This function assumes that H: [0, 360), S: [0, 1], V: [0, 1].
	inline bool eval(const float hsv0[3], const float hsv1[3]);


private:
	EVAL_ORDER eval_order;
	FUNC_TYPE h_func;
	FUNC_TYPE s_func;
	FUNC_TYPE v_func;

	float h_breakpoint;


	int p[3];	// permutation
	int (HSVComparator::*comp[3])(float val0, float val1);

	inline int isMin_H(float h0, float h1);													// check if h0 is smaller than h1
	inline int isMax_H(float h0, float h1);													// check if h0 is larger than h1
	inline int isWarm1_H(float h0, float h1);												// check if h0 is warmer than h1
	inline int isCool1_H(float h0, float h1);												// check if h0 is cooler than h1
	inline int isWarm2_H(float h0, float h1);												// check if h0 is warmer than h1
	inline int isCool2_H(float h0, float h1);												// check if h0 is cooler than h1
	inline int isMin_S(float s0, float s1) { return (s0 < s1) ? 1 : ((s0 > s1) ? -1 : 0); }
	inline int isMax_S(float s0, float s1) { return (s0 > s1) ? 1 : ((s0 < s1) ? -1 : 0); }
	inline int isMin_V(float v0, float v1) { return (v0 < v1) ? 1 : ((v0 > v1) ? -1 : 0); }
	inline int isMax_V(float v0, float v1) { return (v0 > v1) ? 1 : ((v0 < v1) ? -1 : 0); }

	inline void clamp1(float &hue);
	inline void clamp2(float &hue, bool &flipped);
};




inline void HSVComparator::setEvalOrder(HSVComparator::EVAL_ORDER order) { 
	eval_order = order;

	switch(order) {
		case HSV:
		default:
			p[0] = 0;
			p[1] = 1;
			p[2] = 2;
			break;

		case SVH:
			p[0] = 1;
			p[1] = 2;
			p[2] = 0;
			break;

		case VHS:
			p[0] = 2;
			p[1] = 0;
			p[2] = 1;
			break;

		case VSH:
			p[0] = 2;
			p[1] = 1;
			p[2] = 0;
			break;

		case HVS:
			p[0] = 0;
			p[1] = 2;
			p[2] = 1;
			break;

		case SHV:
			p[0] = 1;
			p[1] = 0;
			p[2] = 2;
			break;
	}
}


inline bool HSVComparator::eval(const float hsv0[3], const float hsv1[3]) {
	int result[2];

	if((result[0] = (this->*comp[p[0]])(hsv0[p[0]], hsv1[p[0]])) == 1) return true;
	else if(result[0] == 0 && (result[1] = (this->*comp[p[1]])(hsv0[p[1]], hsv1[p[1]])) == 1) return true;
	else if(result[0] == 0 && result[1] == 0 && (this->*comp[p[2]])(hsv0[p[2]], hsv1[p[2]]) == 1) return true;
	else return false;
}


inline void HSVComparator::setEvalFunc_H(HSVComparator::FUNC_TYPE func) {
	switch(func) {
		case MIN:
			comp[0] = &HSVComparator::isMin_H;
			break;

		case MAX:
			comp[0] = &HSVComparator::isMax_H;
			break;

		case WARM1:
			comp[0] = &HSVComparator::isWarm1_H;
			break;

		case COOL1:
			comp[0] = &HSVComparator::isCool1_H;
			break;

		case WARM2:
			comp[0] = &HSVComparator::isWarm2_H;
			break;

		case COOL2:
			comp[0] = &HSVComparator::isCool2_H;
			break;

		default:
			return;
	}

	h_func = func;
}


inline void HSVComparator::setEvalFunc_S(HSVComparator::FUNC_TYPE func) {
	switch(func) {
		case MIN:
			comp[1] = &HSVComparator::isMin_S;
			break;

		case MAX:
			comp[1] = &HSVComparator::isMax_S;
			break;

		default:
			return;
	}

	s_func = func;
}


inline void HSVComparator::setEvalFunc_V(HSVComparator::FUNC_TYPE func) {
	switch(func) {
		case MIN:
			comp[2] = &HSVComparator::isMin_V;
			break;

		case MAX:
			comp[2] = &HSVComparator::isMax_V;
			break;

		default:
			return;
	}

	v_func = func;
}


inline void HSVComparator::setHueBreakPoint(float val) {
	if(val >= 360.f) {
		while(val >= 360.f) val -= 360.f;
	}
	else if(val < 0) {
		while(val < 0) val += 360.f;
	}

	h_breakpoint = val;
}


// compute a value relative to a hue breakpoint. Clamped if a value is out of range [0, 360)
// both input and a breakpoint assumed to be in a range [0, 360)
inline void HSVComparator::clamp1(float &hue) {
	hue -= h_breakpoint;

	if(hue < 0) hue += 360.f;
}


// compute a value relative to a hue breakpoint. Clamped if a value is out of range [0, 180]
// both input and a breakpoint assumed to be in a range [0, 360)
inline void HSVComparator::clamp2(float &hue, bool &flipped) {
	clamp1(hue);

	if(hue > 180.f) {
		flipped = true;
		hue = 360.f - hue;
	}
	else {
		flipped = false;
	}
}


inline int HSVComparator::isMin_H(float h0, float h1) {
	if(h0 == -1.f) return -1;
	else if(h1 == -1.f) return 1;

	clamp1(h0);
	clamp1(h1);

	return (h0 < h1) ? 1 : ((h0 > h1) ? -1 : 0);
}


inline int HSVComparator::isMax_H(float h0, float h1) {
	if(h0 == -1.f) return -1;
	else if(h1 == -1.f) return 1;

	clamp1(h0);
	clamp1(h1);

	return (h0 > h1) ? 1 : ((h0 < h1) ? -1 : 0);
}


inline int HSVComparator::isWarm1_H(float h0, float h1) {
	bool flip[2];

	if(h0 == -1.f) return -1;
	else if(h1 == -1.f) return 1;

	clamp2(h0, flip[0]);
	clamp2(h1, flip[1]);

	if(h0 < h1) return 1;
	else if(h1 < h0) return -1;
	else if(flip[0] == false && flip[1] == true) return 1;
	else if(flip[0] == true && flip[1] == false) return -1;
	else return 0;
}


inline int HSVComparator::isCool1_H(float h0, float h1) {
	bool flip[2];

	if(h0 == -1.f) return -1;
	else if(h1 == -1.f) return 1;

	clamp2(h0, flip[0]);
	clamp2(h1, flip[1]);

	if(h0 > h1) return 1;
	else if(h1 > h0) return -1;
	else if(flip[0] == false && flip[1] == true) return 1;
	else if(flip[0] == true && flip[1] == false) return -1;
	else return 0;
}


inline int HSVComparator::isWarm2_H(float h0, float h1) {
	bool flip[2];

	if(h0 == -1.f) return -1;
	else if(h1 == -1.f) return 1;

	clamp2(h0, flip[0]);
	clamp2(h1, flip[1]);

	if(h0 < h1) return 1;
	else if(h1 < h0) return -1;
	else if(flip[0] == true && flip[1] == false) return 1;
	else if(flip[0] == false && flip[1] == true) return -1;
	else return 0;
}


inline int HSVComparator::isCool2_H(float h0, float h1) {
	bool flip[2];

	if(h0 == -1.f) return -1;
	else if(h1 == -1.f) return 1;

	clamp2(h0, flip[0]);
	clamp2(h1, flip[1]);

	if(h0 > h1) return 1;
	else if(h1 > h0) return -1;
	else if(flip[0] == true && flip[1] == false) return 1;
	else if(flip[0] == false && flip[1] == true) return -1;
	else return 0;
}

