#include "StdAfx.h"
#include "RegionTransformation.h"

RegionTransformation::RegionTransformation(void)
: imgdat(NULL)
, DatImage(NULL)
, DatImageDefined(false)
, _img(NULL)
{
}

RegionTransformation::~RegionTransformation(void)
{
	cvReleaseImage(&_img);
	releaseImageDat(&imgdat);
}

/*
 ***********************************************************************************
 *	getTranfPixel function
 *	----------------------
 *	-> apply a transformation h to a pixel p
 *
 *	INPUT
 *	p:			a pixel (x,y), x,y of type float
 *	h:			a mapping|transformation matrix
 *
 *	OUTPUT
 *	p' = h(p)	a pixel (x,y), x,y of type float
 ***********************************************************************************
 */
CvPoint2D32f RegionTransformation::getTranfPixel(CvPoint2D32f p, double* h)
{
	CvPoint2D32f pnext;									/* transformed pixel */
	double w = h[6]*p.x + h[7]*p.y + h[8];				/* normalizing factor */
	pnext.x = (float)((h[0]*p.x + h[1]*p.y + h[2])/w);
	pnext.y = (float)((h[3]*p.x + h[4]*p.y + h[5])/w);

	return pnext;
}
/*
 ***********************************************************************************
 *	getIdCol function
 *	-----------------
 *	-> color code a region id
 *
 *	INPUT
 *	id:			a region id
 *
 *	OUTPUT
 *	(r, g, b)	a pixel color, each channel of type double
 ***********************************************************************************
 */
CvScalar RegionTransformation::getIdCol(int id)
{
	return CV_RGB(63*(id%4 + 1),85*(id%3+1),255*(id%2));
}

/*
 ***********************************************************************************
 *	displayVec function
 *	-------------------					  -->
 *	-> draw a arrow representing a vector pq
 *
 *	INPUT
 *	img:			where to draw
 *	p:				starting point (xs,ys), xs,ys of type int
 *	q:				ending point (xe,ye), xe,ye of type int
 *	line_color:		color of the drawn lines
 *	line_thickness:	thickness of the drawn lines
 *	dispArrow:		display the arrow if 1, else display just a line
 *
 *	OUTPUT
 *	none
 ***********************************************************************************
 */
void RegionTransformation::displayVec(IplImage* img, CvPoint p, CvPoint q, CvScalar line_color, int line_thickness, int dispArrow)
{
	double angle;
	double hypotenuse;
	static const double pi = 3.14159265358979323846;

	cvLine( img, p, q, line_color, line_thickness, CV_AA, 0 );
		
	if(dispArrow)
	{
		angle = atan2( (double) p.y - q.y, (double) p.x - q.x );
		hypotenuse = sqrt( (double)(p.y - q.y)*(p.y - q.y) + (double)(p.x - q.x)*(p.x - q.x) );
		p.x = (int) (q.x + 9 * cos(angle + pi / 4));
		p.y = (int) (q.y + 9 * sin(angle + pi / 4));
		cvLine( img, p, q, line_color, line_thickness, CV_AA, 0 );
		p.x = (int) (q.x + 9 * cos(angle - pi / 4));
		p.y = (int) (q.y + 9 * sin(angle - pi / 4));
		cvLine( img, p, q, line_color, line_thickness, CV_AA, 0 );
	}
}

/*
 ***********************************************************************************
 *	loadDat function
 *	----------------
 *	-> create an image data structure and load the data from a file into it
 *
 *	INPUT
 *	filename:		name of the file
 *
 *	OUTPUT
 *	image data structure containing the loaded data
 ***********************************************************************************
 */
bool RegionTransformation::loadDat(const char* filename) {
	FILE* fp = NULL;
	regionDat* rIt;
	int* pIt;
	//imageDat* imgdat = NULL;
	char *labels = NULL;
	char *lIt;
	double cosTheta,sinTheta,cx,cy,cosTheta2,sinTheta2,cx2,cy2,sx,sy;
	int nBytesPerId;


	/* open the file */
	fp = fopen(filename,"rb");

	if(fp == NULL) return false;


	// make sure we release memory allocated to imgdat
	releaseImageDat(&imgdat);

	imgdat = (imageDat*)calloc(1,sizeof(imageDat));


	/* load the width, height and maxId */
	if(fscanf(fp,"%d %d %d\n",&(imgdat->width),&(imgdat->height),&(imgdat->nRegions))<3)
	{
		fprintf(stderr,"In loadDat: cannot read parameters\n");
		fclose(fp);
		free(imgdat);
		imgdat = NULL;

		return false;
	}

	/* allocate memory */
	nBytesPerId = (int)(log((double)imgdat->nRegions)/(8*log(2.0))) + 1;
	imgdat->regions = (regionDat*)calloc(imgdat->nRegions,sizeof(regionDat));
	pIt = imgdat->labels = (int*)calloc(imgdat->width*imgdat->height,sizeof(int));
	lIt = labels = (char*)calloc(imgdat->width*imgdat->height,nBytesPerId*sizeof(char));

	/* load each region information */
	for(rIt = imgdat->regions;rIt < imgdat->regions + imgdat->nRegions;rIt++)
	{
		if(	fscanf(fp,"%d	%lf %lf %lf %lf %lf %d %lf %lf %lf %lf %lf %d ",
						&(rIt->id),
						&(rIt->eCur.x),&(rIt->eCur.y),
						&(rIt->eCur.majorAxis),&(rIt->eCur.minorAxis),
						&(rIt->eCur.theta),&(rIt->eCur.nPixels),
						&(rIt->eNext.x),&(rIt->eNext.y),
						&(rIt->eNext.majorAxis),&(rIt->eNext.minorAxis),
						&(rIt->eNext.theta),&(rIt->eNext.nPixels))< 13 ||

			fscanf(fp,"%lf %lf %lf %lf %lf %lf %lf %lf %lf\n",
						&(rIt->h[0]),&(rIt->h[1]),&(rIt->h[2]),
						&(rIt->h[3]),&(rIt->h[4]),&(rIt->h[5]),
						&(rIt->h[6]),&(rIt->h[7]),&(rIt->h[8])) < 9)
		{
			fprintf(stderr,"In loadDat: cannot read parameters\n");
			free(imgdat->regions);
			free(imgdat->labels);
			fclose(fp);
			free(imgdat);
			free(labels);
			imgdat = NULL;

			return false;
		}

		/* fill the ellipse transformation */
		if(rIt->eCur.nPixels>0 && rIt->eNext.nPixels>0)
		{
			rIt->eTransf.tx = rIt->eNext.x - rIt->eCur.x;
			rIt->eTransf.ty = rIt->eNext.y - rIt->eCur.y;
			rIt->eTransf.sM = rIt->eNext.majorAxis/rIt->eCur.majorAxis;
			rIt->eTransf.sm = rIt->eNext.minorAxis/rIt->eCur.minorAxis;
			rIt->eTransf.dTheta = rIt->eNext.theta - rIt->eCur.theta;

			/* threshold if ellipses close to circle and if huge angle difference */
			if(((rIt->eCur.minorAxis/rIt->eCur.majorAxis)>0.9 || (rIt->eNext.minorAxis/rIt->eNext.majorAxis)>0.9) && fabs(rIt->eTransf.dTheta) > 45)
			{
				rIt->eTransf.dTheta/=45;
			}
			rIt->eTransf.s = rIt->eNext.nPixels / rIt->eCur.nPixels;
			sx = rIt->eTransf.sM;							sy = rIt->eTransf.sm;
			cx = rIt->eCur.x;								cy = rIt->eCur.y;
			cosTheta = cos(M_PI*rIt->eCur.theta/180.);		sinTheta = sin(M_PI*rIt->eCur.theta/180.);
			cx2 = rIt->eNext.x;								cy2 = rIt->eNext.y;
			cosTheta2 = cos(M_PI*(rIt->eCur.theta + rIt->eTransf.dTheta)/180.);
			sinTheta2 = sin(M_PI*(rIt->eCur.theta + rIt->eTransf.dTheta)/180.);

			rIt->eTransf.h[0] = sx*cosTheta2*cosTheta + sinTheta2*sinTheta;		rIt->eTransf.h[1] = sx*cosTheta2*sinTheta - sinTheta2*cosTheta;		rIt->eTransf.h[2] = cx2 - sx*cosTheta2*(cx*cosTheta + cy*sinTheta) - sinTheta2*(cx*sinTheta - cy*cosTheta);
			rIt->eTransf.h[3] = sinTheta2*cosTheta - sy*cosTheta2*sinTheta;		rIt->eTransf.h[4] = sinTheta2*sinTheta + sy*cosTheta2*cosTheta;		rIt->eTransf.h[5] = cy2 - sinTheta2*(cx*cosTheta + cy*sinTheta) + sy*cosTheta2*(cx*sinTheta - cy*cosTheta);
			rIt->eTransf.h[6] = 0;												rIt->eTransf.h[7] = 0;												rIt->eTransf.h[8] = 1;
		}
		else
		{
			memset(&(rIt->eTransf),0,sizeof(ellipseTransf));
			rIt->eTransf.sm = rIt->eTransf.sM = rIt->eTransf.s = 1;
			rIt->eTransf.h[0] = rIt->eTransf.h[4] = rIt->eTransf.h[8] = 1;
		}
	}

	/* get the labels */
	if(fread(labels,nBytesPerId*sizeof(char),imgdat->width*imgdat->height,fp) != imgdat->width*imgdat->height)
	{
		fprintf(stderr,"In loadDat: cannot read data from id map file %s\n",filename);
		free(labels);
		free(imgdat->regions);
		free(imgdat->labels);
		fclose(fp);
		free(imgdat);
		imgdat = NULL;

		return false;
	}

	for(;pIt<imgdat->labels + imgdat->width*imgdat->height;pIt++)
	{
		memcpy(pIt,lIt,nBytesPerId*sizeof(char));
		if(*pIt<=0 || *pIt>imgdat->nRegions) *pIt = 0;
		lIt += nBytesPerId;
	}

	/* free memory and close the file */
	free(labels);

	if(fclose(fp)) {
		fprintf(stderr,"In loadDat: cannot close the dat file %s\n",filename);

		return false;
	}

	return true;
}

/*
 ***********************************************************************************
 *	releaseImageDat function
 *	------------------------
 *	-> release a image data structure
 *
 *	INPUT
 *	imDat:		address of the structure
 *
 *	OUTPUT
 *	1 if released
 ***********************************************************************************
 */
int RegionTransformation::releaseImageDat(imageDat** imDat)
{
	if(imDat<=NULL)
	{
		fprintf(stderr,"In releaseImageDat: bad input parameters\n");
		return 0;
	}

	if(*imDat <= NULL){*imDat = NULL; return -1;}

	if((*imDat)->regions > NULL) free((*imDat)->regions);
	if((*imDat)->labels > NULL) free((*imDat)->labels);
	free(*imDat);
	*imDat = NULL;

	return 1;
}

void RegionTransformation::getDat(const char* filename)
{
	IplImage* img = NULL;
	//char* filename = DEFAULT_DAT;
	CvPoint2D32f pnext;
	int step = 20;
	int id = 0;

		
	/* load .dat file */
	loadDat(filename);



	/* display the labels */
	img = cvCreateImage( cvSize(imgdat->width,imgdat->height), IPL_DEPTH_8U, 3 );
	for(int i=0;i<img->height;i++)
	{
		for(int j=0;j<img->width;j++)
		{
			//printf("i: %d j: %d label: %d\n",i,j,imgdat->labels[i*img->width+j]);
			cvSet2D(img,i,j,getIdCol(imgdat->labels[i*img->width+j]));
		}
	}

	/* display the flow for K to K+1 */
	for(int i=0;i<img->height;i+=step)
	{
		for(int j=0;j<img->width;j+=step)
		{
			id = imgdat->labels[i*img->width+j]-1;
			if(id<0) continue;
			pnext = getTranfPixel(cvPoint2D32f((float)j,(float)i),imgdat->regions[id].h);
			displayVec(img,cvPoint(j,i),cvPoint((int)pnext.x,(int)pnext.y),cvScalarAll(0),1,1);
		}
	}

	if(DatImageDefined == true) {
		DatImage = NULL;
		cvReleaseImage(&_img);
		_img = img;
		DatImage = (unsigned char *)_img->imageDataOrigin;
	}

	DatImageDefined = true;
}


int RegionTransformation::getID(int x, int y) {
	if(0 <= x && x < imgdat->width && 0 <= y && y < imgdat->height) {
		return imgdat->labels[x + imgdat->width * y];
	}
	else {
		return -1;
	}
}

