/*

The file contains routines for a priority-list scheme described in
the following article:

  William Pugh, "Skip Lists: A Probabilistic Alternative to Balanced Trees",
  Communications of the ACM, Vol. 33, No. 6, June 1990.

The basic idea is a variant on a simple linked-list.  Some of the list
elements have more than one forward pointer, and these additional pointers
point to increasingly more distant elements in the list.

Greg Turk, May 1995

*/

#pragma warning (once: 4311)

#include "StdAfx.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "skiplist.h"

static Skip_element *update[100];


/******************************************************************************
Initialize a sorted list.

Entry:
  compare_proc - function with which to compare list elements (for sorting)
  max_lev      - maximum number of pointer levels to use in skip-list

Exit:
  returns the new list
******************************************************************************/

SkipList::SkipList(int (*compare_proc)(char *, char *), int max_lev)
{
  head.level = 0;
  compare = compare_proc;
  num_elements = 0;
  max_level = max_lev;

  /* allocate forward pointers for the list header */
  head.forward = (Skip_element **)
    malloc (sizeof (Skip_element *) * max_level);

  /* set the skip-list to NULL */
  for (int i = 0; i < max_level; i++)
    head.forward[i] = NULL;
}


/******************************************************************************
Pick a random level for an element in a skip-list.
******************************************************************************/

static int random_level(int max_level)
{
  int level;

  level = 0;
  while (rand() & 1)
    level++;

  if (level >= max_level)
    return (max_level - 1);
  else
    return (level);
}


/******************************************************************************
Add an element to a skip-list.

Entry:
  data - data for new element
******************************************************************************/

void SkipList::add(char *data)
{
  Skip_element *ptr;
  Skip_element *elem;
  int newlevel;

  /* look for the position for the new element in sorted skip-list */
  /* note: (compare) is a function invocation to compare two elements */

  ptr = &head;
  for (int i = head.level; i >= 0; i--) {
    while (ptr->forward[i] != NULL &&
	   (compare) (ptr->forward[i]->data, data) < 0) {
      ptr = ptr->forward[i];
    }
    while (ptr->forward[i] != NULL &&
	   (compare) (ptr->forward[i]->data, data) == 0 &&
	   ((int) ptr->forward[i]->data) < ((int) data)) {
      ptr = ptr->forward[i];
    }
    update[i] = ptr;
  }

  /* we need to add a new element, so pick the number */
  /* of forward pointers that the new element will have */
  newlevel = random_level (max_level);

  /* if this level is larger than the current level, augment the update list */
  if (newlevel > head.level) {
    for (int i = head.level + 1; i <= newlevel; i++)
      update[i] = &head;
    head.level = newlevel;
  }

  /* make new element for list */

  elem = (Skip_element *) malloc (sizeof (Skip_element));
  elem->level = newlevel;
  elem->forward = (Skip_element **)
    malloc (sizeof (Skip_element *) * (newlevel + 1));
  elem->data = data;

  /* perform the insertion */
  for (int i = 0; i <= newlevel; i++) {
    elem->forward[i] = update[i]->forward[i];
    update[i]->forward[i] = elem;
  }

  /* count the new element */
  num_elements++;
}


/******************************************************************************
Remove an element from a skip-list.  Warning: Deletes the skip-list element,
but not the data contained in it.

Entry:
  list - skip-list to remove from
  data - data to remove from list

Exit:
  returns 0 if removed successfully, 1 if element not present in list (error)
******************************************************************************/

int SkipList::remove(char *data)
{
  Skip_element *ptr;

  /* look for the position for the element in sorted skip-list */
  /* note: (compare) is a function invocation to compare two elements */

  ptr = &head;
  for (int i = head.level; i >= 0; i--) {
    while (ptr->forward[i] != NULL &&
	   (compare) (ptr->forward[i]->data, data) < 0) {
      ptr = ptr->forward[i];
    }
    while (ptr->forward[i] != NULL &&
	   (compare) (ptr->forward[i]->data, data) == 0 &&
	   ((int) ptr->forward[i]->data) < ((int) data)) {
      ptr = ptr->forward[i];
    }
    update[i] = ptr;
  }

  /* see if we've found the element to remove */
  if (ptr->forward[0]->data == data) {
    ptr = ptr->forward[0];
  }
  else {
    /* if we get here, elem not in list so we should return an error code */
    return (1);
  }

  /* remove the element */
  for (int i = 0; i <= ptr->level; i++)
    update[i]->forward[i] = ptr->forward[i];

  free (ptr->forward);
  free (ptr);

  /* decrease the element count of the list */
  num_elements--;

  /* see if we need to decrease the number of levels in the header */
  for (int i = head.level; i > 0; i--) {
    if (head.forward[i] == NULL)
      head.level--;
    else
      break;
  }

  /* return code saying elem was removed successfully */
  return (0);
}


/******************************************************************************
Find an element that matches the given data, according to the comparison
routine for the list.

Entry:
  list - skip-list to search
  data - data to match

Exit:
  returns pointer to matching data if successful, NULL if not
******************************************************************************/

char *SkipList::find(char *data)
{
  Skip_element *ptr;

  /* look for the position for the element in sorted skip-list */
  /* note: (compare) is a function invocation to compare two elements */

  ptr = &head;
  for (int i = head.level; i >= 0; i--) {
    while (ptr->forward[i] != NULL &&
	   (compare) (ptr->forward[i]->data, data) < 0) {
      ptr = ptr->forward[i];
    }
    while (ptr->forward[i] != NULL &&
	   (compare) (ptr->forward[i]->data, data) == 0 &&
	   ((int) ptr->forward[i]->data) < ((int) data)) {
      ptr = ptr->forward[i];
    }
  }

  /* see if we've got a match, and return the appropriate info */

  if (ptr->forward[0] == NULL)
    return (NULL);

  if ((compare) (ptr->forward[0]->data, data) == 0)
    return (ptr->forward[0]->data);
  else
    return (NULL);
}


/******************************************************************************
Return the last element in a skip-list.

Exit:
  returns last element of list, or NULL if list is empty
******************************************************************************/

char *SkipList::last()
{
  Skip_element *ptr;

  /* return NULL if the list is empty */

  if (num_elements == 0)
    return (NULL);

  /* look for the last element */

  ptr = &head;
  for (int i = head.level; i >= 0; i--) {
    while (ptr->forward[i] != NULL)
      ptr = ptr->forward[i];
  }

  /* return pointer to last element */
  return ((char *) ptr->data);
}


/******************************************************************************
Return the first element in a skip-list.

Exit:
  returns first element of list, or NULL if the list is empty
******************************************************************************/

char *SkipList::first()
{
  if (head.forward[0] == NULL)
    return (NULL);
  else
    return ((char *) head.forward[0]->data);
}


/******************************************************************************
Return the number of elements in a skip-list.

Entry:
  list - skip-list to return number of elements of

Exit:
  returns number of elements in skip-list
******************************************************************************/

int SkipList::count()
{
  return (num_elements);
}


/******************************************************************************
Print out a skip-list.

Entry:
  list       - list to print
  print_data - routine to print the data in an element
******************************************************************************/

void SkipList::print(void (print_data)(char *))
{
  int count;
  Skip_element *ptr;
  Skip_element *p;

  for (ptr = &head; ptr != NULL; ptr = ptr->forward[0]) {

    if (ptr == &head) {
      printf ("(header) ");
    }
    else {
      (*print_data) (ptr->data);
    }

    printf (" level = %d: ", ptr->level);

    for (int i = 0; i <= ptr->level; i++) {
      count = 0;
      for (p = ptr; p != NULL && p != ptr->forward[i]; p = p->forward[0])
	count++;
      printf ("%5d", count);
    }

    printf ("\n");
  }
}


/******************************************************************************
Free a list.

Entry:
  list      - list to free up
  free_data - flag whether to free data, 1 = free it, 0 = don't
******************************************************************************/

SkipList::~SkipList()
{
#if 0
  Skip_element *elem;
  Skip_element *next;

  /* free up all the elements in the list */
  for (elem = list->head.forward[0]; elem != NULL; elem = next) {
    next = elem->forward[0];
    free (elem->forward);
    if (free_data)
      free (elem->data);
    free (elem);
  }

  /* free up the list info */
  free (list);
#endif
}

