/*

Data structure for creating vector/tensor fields on a planar domain

Eugene Zhang 2003

*/

#ifndef __TENDES_PLANE_H__
#define __TENDES_PLANE_H__


#include "skiplist.h"
#include "ply.h"
#include "icMatrix.H"
#include "icComplex.H"

#include <list>
#include <vector>
using namespace std;


const double EPS = 1.0e-6;
const double PI=3.1415926535898;

extern float mag_ratio;

/* forward declarations */
class Vertex;
class Triangle;
class Polyhedron;
class Edge;
class VertexList;
class EdgeList;
class Corner;
class Singularity;
class SeparatricesList;

class Vertex {
public:
	double x,y,z;
	int index;
	unsigned char temp_use;
	unsigned int first_use, last_use, smallest_length, smallest_start, smallest_end;
	//////////////////////////////////////////////////////////////////////////////////////////////
	 int regionID;
	 bool use_zero_value;		// use zero value or vector/tensor field stored on vec_fld_painting/ten_fld_painting

	//////////////////////////////////////////////////////////////////////////////////////////////

	int on_path;
	int last_touched;
	int status; // -1 = alive, 0 = close, 1 = far
	double dist_temp;
	double value;
	Vertex *parent;

	int nedges;
	int traced;
	int pos_in_matrix;
	double floater_lamda;

	icVector2 vec_fld, vec_fld_painting, old_vec_fld, undo_vec_fld, redo_vec_fld;
	icVector2 div_fld;
	icMatrix2x2 ten_fld, ten_fld_painting, old_ten_fld, undo_ten_fld, redo_ten_fld;

	float stepsize;
	double new_ts[8][3];

	int ntris;
	Triangle **tris;
	int max_tris;

	icVector3 normal;
	EdgeList *adjacent_edges;

	void *other_props;

	Vertex(double xx, double yy, double zz) { 
		x = xx;
		y = yy;
		z = zz;

		tris = NULL;
		adjacent_edges = NULL;
		other_props = NULL;

		regionID = -1;

		stepsize = 0;
		use_zero_value = false;
	}

	friend Edge;
	friend Triangle;
	friend Polyhedron;
};


class Edge {
public:
	int index;
	Vertex *verts[2];
	int ntris;
	Triangle **tris;

	int on_path;
	unsigned char clicked;
	unsigned char temp_use;
	unsigned char used_for_calc_lamba;
	double length;
	unsigned char touched;
	char transverse;
	double floater_lamda[2];


	Edge() {
		for(int i = 0; i < 2; i++) verts[i] = NULL;
		tris = NULL;
	}
};


class Triangle {
public:
	int index;
	int nverts;
	Vertex *verts[3];
	Edge *edges[3];

	int on_path;

	unsigned char visible; 

	unsigned char temp_use;
	double angle[3];
	icVector2 vert_local[3], edge_normal[3];
	double area;
	icVector3 normal;
	icVector2 x_dir, y_dir;
	icMatrix3x3 jacobian;
	double x1, x2, y2;  // assume x0=y0=y1=0.0
	//	double vec_a, vec_b, vec_c, vec_d, vec_e, vec_f;  // assume V(x, y) = (Ax + By + C, Dx + Ey + F)
	//	double ten_a, ten_b, ten_c, ten_d, ten_e, ten_f;  // assume T(x, y) = (Ax + By + C, Dx + Ey + F, Dx + Ey + F, -(Ax + By + C))
	double vecten_a, vecten_b, vecten_c, vecten_d, vecten_e, vecten_f;  // assume V(x, y) = (Ax + By + C, Dx + Ey + F)
	Singularity *the_singularity;
	void *other_props;
	icVector3 center;
	int clicked;
	Corner *corners[3];

public:
	Triangle() {
		for(int i = 0; i < 3; i++) {
			verts[i] = NULL;
			edges[i] = NULL;
			corners[i] = NULL;
		}

		the_singularity = NULL;
		other_props = NULL;
	}

	double plane_equation (double &, double &, double &, double &);

	inline void convert_global_to_local(icVector2 in, icVector2 *out) {
		icVector2 temp = in - icVector2(verts[0]->x, verts[0]->y);
		out->entry[0] = dot(temp, x_dir);
		out->entry[1] = dot(temp, y_dir);
	}
	
	inline void convert_local_to_global(icVector2 in, icVector2 *out) {
		*out = x_dir * in.entry[0] + y_dir * in.entry[1] + icVector2(verts[0]->x, verts[0]->y);
	}
	
	inline void convert_global_to_local_vector(icVector2 in, icVector2 *out) {
		out->entry[0] = dot(in, x_dir);
		out->entry[1] = dot(in, y_dir);
	}
	
	inline void convert_local_to_global_vector(icVector2 in, icVector2 *out) {
		*out = x_dir * in.entry[0] + y_dir * in.entry[1];
	}
	
	inline void convert_global_to_local_tensor(icMatrix2x2 in, icMatrix2x2 *out) {
		icMatrix2x2 trans(x_dir.entry[0], y_dir.entry[0], x_dir.entry[1], y_dir.entry[1]);
		*out = othoconjugate(in, trans);
	}
	
	inline void convert_local_to_global_tensor(icMatrix2x2 in, icMatrix2x2 *out) {
		icMatrix2x2 trans(x_dir.entry[0], y_dir.entry[0], x_dir.entry[1], y_dir.entry[1]);
		*out = othoconjugate(in, transpose(trans));
	}

	inline void get_bary_coor(icVector2 in, double *bary_a, double *bary_b) {
		*bary_b = in.entry[1] / y2;
		*bary_a = (y2*in.entry[0] - x2*in.entry[1])/(x1*y2);
	}
	
	inline int quick_index(int vec_ten_mode) {
		double angle[3], total_angle;
		icVector2 normed_vec[3];
		
		if(vec_ten_mode == 0) {
			for(int i = 0; i < 3; i++) {
				if(verts[i]->vec_fld == 0.0) return -10;  // bad case

				normed_vec[i] = verts[i]->vec_fld;
				normalize(normed_vec[i]);
			}
		}
		else {
			for(int i = 0; i < 3; i++) {
				if(verts[i]->ten_fld == 0.0) return -10;  // bad case

				normed_vec[i].set(verts[i]->ten_fld.entry[0][0], verts[i]->ten_fld.entry[0][1]);
				normalize(normed_vec[i]);
			}
		}
	//if (dot(normed_vec[0], normed_vec[1])>1.0-1.0e-8 && dot(normed_vec[1], normed_vec[2])>1.0-1.0e-8 && dot(normed_vec[0], normed_vec[1])>1.0-1.0e-8)
	//return -10;  // bad case

		total_angle = 0.0;

		for(int i = 0; i < 3; i++) {
			angle[i] = atan2(normed_vec[(i+1) % 3].entry[1], normed_vec[(i+1) % 3].entry[0]) - atan2(normed_vec[i].entry[1], normed_vec[i].entry[0]);
			
			while (angle[i] > PI) angle[i] -= 2*PI;
			while (angle[i] < -PI) angle[i] += 2*PI;

			total_angle += angle[i];
		}

		return (int)floor(total_angle/(2*PI)+0.5);
	}

	friend Vertex;
	friend Polyhedron;
};


class Corner {
public:
	int index;
	Vertex *v;
	Edge *e;
	Corner *p, *n, *o;
	Triangle *t;
	char transverse;

	Corner() {
		v = NULL;
		e = NULL;
		p = NULL;
		n = NULL;
		o = NULL;
		t = NULL;
	}
};


class VertexList {
public:
	int num_verts;
	int max_verts;
	Vertex **verts;

	VertexList (int max) {
		max_verts = max;
		verts = new Vertex *[max_verts];

		for(int i = 0; i < max_verts; i++) verts[i] = NULL;
		num_verts = 0;
	}

/******************************************************************************
Insert a vertex into a vertex list if it isn't already in the list.
******************************************************************************/

	void insert_unique(Vertex *v) {
		/* look for vertex to see if it is already in the list */
		for(int i = 0; i < num_verts; i++) {
			if(verts[i] == v) return;
		}

		append(v);
	}

	void append(Vertex *v) {
		/* if we get here it isn't in the list, so add it */
		/* first make sure there is enough room for new vertex */
		if(num_verts >= max_verts) {
			max_verts += 10;
			Vertex **tlist = new Vertex *[max_verts];
			
			for(int i = 0; i < num_verts; i++) tlist[i] = verts[i];
			for(int i = num_verts; i < max_verts; i++) tlist[i] = NULL;

			delete (verts);
			verts = tlist;
		}

		/* add new vertex to list */
		verts[num_verts] = v;
		num_verts++;
	}

	void finalize() {
		if(this != NULL) {
			delete(verts);
			delete(this);
		}
	}
};


class TriangleList {
public:
	int num_tris;
	int max_tris;
	Triangle **tris;

	TriangleList(int max) {
		max_tris = max;
		tris = new Triangle *[max_tris];

		for(int i = 0; i < max_tris; i++) tris[i] = NULL;
		num_tris = 0;
	}

	void insert_unique(Triangle *t) {
		/* look for vertex to see if it is already in the list */
		for(int i = 0; i < num_tris; i++) {
			if(tris[i] == t) return;
		}

		append(t);
	}

	void append(Triangle *t) {
		if(num_tris >= max_tris) {
			max_tris += 10;
			Triangle **tlist = new Triangle *[max_tris];

			for(int i = 0; i < num_tris; i++) tlist[i] = tris[i];
			for(int i = num_tris; i < max_tris; i++) tlist[i] = NULL;

			//delete (tris);
			delete[] tris;
			tris = tlist;
		}

		tris[num_tris] = t;
		num_tris++;
	}

	void finalize() {
		if (this != NULL) {
			//delete(tris);
			delete[] tris;
			delete(this);
		}
	}
};

class EdgeList {
public:
	int num_edges;
	int max_edges;
	Edge **edges;

	EdgeList (int max) {
		max_edges = max;
		edges = new Edge *[max_edges];

		for(int i = 0; i < max_edges; i++) edges[i] = NULL;
		num_edges = 0;
	}

	void insert_unique(Edge *e) {
		/* look for vertex to see if it is already in the list */
		for(int i = 0; i < num_edges; i++) {
			if(edges[i] == e) return;
		}

		append(e);
	}

	void append(Edge *e) {
		/* look for vertex to see if it is already in the list */
		for(int i = 0; i < num_edges; i++) {
			if(edges[i] == e) return;
		}

		/* if we get here it isn't in the list, so add it */

		/* first make sure there is enough room for new vertex */
		if(num_edges >= max_edges) {
			max_edges += 10;
			Edge **tlist = new Edge *[max_edges];
			
			for(int i = 0; i < num_edges; i++) tlist[i] = edges[i];
			for(int i = num_edges; i < max_edges; i++) tlist[i] = NULL;

			delete[] edges;
			edges = tlist;
		}

		/* add new edge to list */

		edges[num_edges] = e;
		num_edges++;
	}

	void finalize() {
		if(this != NULL) {
			delete[] edges;
			delete(this);
		}
	}
};


class CornerList {
public:
	int num_corners;
	int max_corners;
	Corner **corners;

	CornerList(int max) {
		max_corners = max;
		corners = new Corner *[max_corners];

		for(int i = 0; i < max_corners; i++) corners[i] = NULL;
		num_corners = 0;
	}

	void append(Corner *c){
		if(num_corners >= max_corners) {
			max_corners += 200;
			Corner **clist = new Corner *[max_corners];

			for(int i = 0; i < num_corners; i++) clist[i] = corners[i];
			for(int i = num_corners; i < max_corners; i++) clist[i] = NULL;
			
			delete(corners);
			corners = clist;
		}

		/* add new Corner to list */
		corners[num_corners] = c;
		num_corners++;
	}

	void insert_unique(Corner *c){
		for(int i = 0; i < num_corners; i++) {
			if(corners[i] == c) return;
		}

		append(c);
	}

	void finalize() {
		if(this != NULL) {
			delete(corners);
			delete(this);
		}
	}
};


class SingularElement {
public:
	double x,y;
	int index;
	unsigned char type;  // 0 = min, 1 = max, 2 = saddle, 3 = center	
	unsigned char order;  // 0 = first order, 1 = second order
	double max_mag, min_mag;  // magnitude along max, min directions
	icVector2 max_dir, min_dir;
	icMatrix2x2 mat, rot, rot_inv, mat_second[2];
	icVector2 center, rotate_center, stretch_center[8];
	int which_mode; // 0 = vecfld, 1 = tenfld
	const SingularElement *ref;


	int regionID;
	int frameID;

	SingularElement(int vec_ten_mode, unsigned char the_order, unsigned char the_type, double xx, double yy, int region_id = 0, int frame_id = 0)
	{ 
		which_mode = vec_ten_mode;
		order = the_order;
		type = the_type; 
		x = xx; 
		y = yy; 
		max_dir.set(1.0, 0.0);
		min_dir.set(0.0, 1.0);
		min_mag = 1.0;
		max_mag = 1.0;
		Construct_mat();
		regionID = region_id;
		frameID = frame_id;
		ref = this;

		calc_display_attributes();
	}


	SingularElement(const SingularElement &se) {
		x = se.x;
		y = se.y;
		index = se.index;
		type = se.type;
		order = se.order;
		max_mag = se.max_mag;
		min_mag = se.min_mag;
		max_dir = se.max_dir;
		min_dir = se.min_dir;
		mat = se.mat;
		rot = se.rot;
		rot_inv = se.rot_inv;

		for(int i = 0; i < 2; i++) mat_second[i] = se.mat_second[i];

		center = se.center;
		rotate_center = se.rotate_center;

		for(int i = 0; i < 8; i++) stretch_center[i] = se.stretch_center[i];

		which_mode = se.which_mode;

		regionID = se.regionID;
		frameID = se.frameID;
		ref = se.ref;
	}


	void Construct_mat() {
		if(order == 0) {
			switch (type) {
				case 0: // min
					mat.set(-max_mag, 0.0, 0.0, -min_mag);
					break;

				case 1: // max
					mat.set(max_mag, 0.0, 0.0, min_mag);
					break;

				case 2: // saddle
					mat.set(max_mag, 0.0, 0.0, -min_mag);
					break;

				case 3: // ccw center
					mat.set(0.0, -max_mag, min_mag, 0.0);
					break;

				case 4: // ccw center
					mat.set(0.0, max_mag, -min_mag, 0.0);
					break;
			}

			if(which_mode == 0) {
				icMatrix2x2 temp(max_dir, min_dir);
				mat.rightMultiply(temp);
				mat.leftMultiply(transpose(temp));
			}
			else {
				double angle = -atan2(max_dir.entry[1], max_dir.entry[0]);
				double the_c = cos(angle), the_s = sin(angle);

				rot.set(the_c, -the_s, the_s, the_c);
				rot_inv.set(the_c, the_s, -the_s, the_c);
				mat.rightMultiply(rot);
			}
		}
		else {
			if(which_mode == 0) {
				double angle = -atan2(max_dir.entry[1], max_dir.entry[0]);
				double the_c = cos(angle), the_s = sin(angle);
				double cc = the_c * the_c, cs = the_c * the_s, ss = the_s * the_s;
				double MM = max_mag * max_mag, Mm = max_mag * min_mag, mm = min_mag * min_mag;

				rot.set(the_c, -the_s, the_s, the_c);
				rot_inv.set(the_c, the_s, -the_s, the_c);

				if(type == 0) {   // dipole 1: note this one behaves differently from that on a tensor field 
					mat_second[0].set(mm, 0.0, 0.0, -MM);
					mat_second[1].set(0.0, Mm, Mm, 0.0);
				}
				else if (type == 1) {  // dipole 2
					mat_second[0].set(-mm, 0.0, 0.0, MM);
					mat_second[1].set(0.0, -Mm, -Mm, 0.0);
				}
				else if (type == 2) {  // monkey saddle
					mat_second[0].set(MM, 0.0, 0.0, -mm);
					mat_second[1].set(0.0, -Mm, -Mm, 0.0);
				}

				for(int pp = 0; pp < 2; pp++) {
					mat_second[pp].rightMultiply(rot);
					mat_second[pp].leftMultiply(rot_inv);
				}
			}
			else {
				double angle = -atan2(max_dir.entry[1], max_dir.entry[0]);
				double the_c = cos(angle), the_s = sin(angle);
				double cc = the_c * the_c, cs = the_c * the_s, ss = the_s * the_s;
				double MM = max_mag * max_mag, Mm = max_mag * min_mag, mm = min_mag * min_mag;

				rot.set(the_c, -the_s, the_s, the_c);
				rot_inv.set(the_c, the_s, -the_s, the_c);

				if(type == 0) {			// dipole 1
					mat_second[0].set(MM, 0.0, 0.0, -mm);
					mat_second[1].set(0.0, Mm, Mm, 0.0);
				}
				else if(type == 1) {	// dipole 2
					mat_second[0].set(-mm, 0.0, 0.0, MM);
					mat_second[1].set(0.0, -Mm, -Mm, 0.0);
				}
				else if(type == 2) {	// monkey saddle
					mat_second[0].set(MM, 0.0, 0.0, -mm);
					mat_second[1].set(0.0, -Mm, -Mm, 0.0);
				}

				for(int pp = 0; pp < 2; pp++) {
					mat_second[pp].rightMultiply(rot);
					mat_second[pp].leftMultiply(rot_inv);
				}
			}
		}
	}

	void calc_display_attributes() {
		center.set(x, y);
		stretch_center[0] = center + mag_ratio*max_mag*max_dir;
		stretch_center[1] = 2.0*center - stretch_center[0];
		stretch_center[2] = center + mag_ratio*min_mag*min_dir;
		stretch_center[3] = 2.0*center - stretch_center[2];
		stretch_center[4] = stretch_center[0] + stretch_center[2] - center;
		stretch_center[5] = 2.0*stretch_center[2] - stretch_center[4];
		stretch_center[6] = 2.0*center - stretch_center[4];
		stretch_center[7] = 2.0*center - stretch_center[5];
		rotate_center = stretch_center[2] + 0.1/length(stretch_center[2] - center)*(stretch_center[2] - center);
	}

	void Update_Position(double xx, double yy, double zz) {
		x = xx; 
		y = yy;
		calc_display_attributes();
	}

	void Update_Angle (double angle) {
		max_dir.set(cos(angle), sin(angle));
		min_dir.set(-sin(angle), cos(angle));
		Construct_mat();
		calc_display_attributes();
	}

	void Update_Magnitude(double max_magnitude, double min_magnitude) {
		max_mag = max_magnitude;
		min_mag = min_magnitude;
		Construct_mat();
		calc_display_attributes();
	}
};


class SingularElementList {
public:
	int num_singularelements;
	int max_singularelements;
	SingularElement **singularelements;

	SingularElementList (int max) {
		max_singularelements = max;
		singularelements = new SingularElement *[max_singularelements];

		for(int i = 0; i < max_singularelements; i++) singularelements[i] = NULL;
		num_singularelements = 0;
	}

	SingularElementList(const SingularElementList &list) {
		num_singularelements = list.num_singularelements;
		max_singularelements = list.max_singularelements;
		singularelements = new SingularElement *[max_singularelements];

		for(int i = 0; i < num_singularelements; i++) {
			singularelements[i] = new SingularElement(*(list.singularelements[i]));
		}

		for(int i = num_singularelements; i < max_singularelements; i++) {
			singularelements[i] = NULL;
		}
	}

	void append(SingularElement *se) {
		if(num_singularelements >= max_singularelements) {
			max_singularelements += 20;
			SingularElement **selist = new SingularElement *[max_singularelements];
			
			for(int i = 0; i < num_singularelements; i++) selist[i] = singularelements[i];
			for(int i = num_singularelements; i < max_singularelements; i++) selist[i] = NULL;

			delete (singularelements);
			singularelements = selist;
		}

		/* add new SingularElement to list */

		singularelements[num_singularelements] = se;
		num_singularelements++;
	}

	void finalize() {
		if(this != NULL) {
			for(int i = 0; i < num_singularelements; i++) {
				if(singularelements[i]) delete(singularelements[i]);
			}

			delete(singularelements);
			delete(this);
		}
	}
};

class RegularElement {
public:
	double x,y;
	double vx, vy, vz;
	//int index;
	icVector2 start_center, end_center, other_end_center, side_pt[4];
	double strength;

	int regionID;
	int frameID;
	const RegularElement *ref;

public:
	RegularElement(double xx, double yy, int region_id = 0, int frame_id = 0) { 
		x = xx; 

		y = yy; 
		vx = 0.0;
		vy = 0.0;
		strength = 1.0;
		calc_display_attributes();

		regionID = region_id;
		frameID = frame_id;
		ref = this;
	}

	RegularElement(const RegularElement &re) {
		x = re.x;
		y = re.y;
		vx = re.vx;
		vy = re.vy;
		regionID=re.regionID;
		start_center = re.start_center;
		end_center = re.end_center;
		other_end_center = re.other_end_center;

		for(int i = 0; i < 4; i++) side_pt[i] = re.side_pt[i];

		strength = re.strength;

		regionID = re.regionID;
		frameID = re.frameID;
		ref = re.ref;
	}

	void calc_display_attributes() {
		start_center.set(x, y);
		end_center = start_center + mag_ratio*icVector2(vx, vy);
		other_end_center = 2.0*start_center - end_center;

		icVector2 x_dir, y_dir, arrow_dir;
		x_dir.set(-vx, -vy);
		y_dir.set(-vy, vx);
		normalize(x_dir);
		normalize(y_dir);
		arrow_dir = 0.5*mag_ratio*(sqrt(3.0)/2.0*x_dir + 0.5*y_dir);
		side_pt[0] = end_center + arrow_dir;
		arrow_dir -= 0.5*mag_ratio * y_dir;
		side_pt[1] = end_center + arrow_dir;
		side_pt[2] = 2.0*start_center - side_pt[0];
		side_pt[3] = 2.0*start_center - side_pt[1];
	}

	void Update_Start_Position(double xx, double yy) {
		x = xx; 
		y = yy;
		calc_display_attributes();
	}
	
	void Update_End_Position (double xx, double yy) {
		double inverse_ratio = 1.0/mag_ratio;
		vx = (xx - x)*inverse_ratio; 
		vy = (yy - y)*inverse_ratio;
		strength = length(icVector2(vx, vy));
		calc_display_attributes();
	}
};


class RegularElementList {
public:
	int num_regularelements;
	int max_regularelements;
	RegularElement **regularelements;

	RegularElementList (int max) {
		max_regularelements = max;
		regularelements = new RegularElement *[max_regularelements];

		for(int i = 0; i < max_regularelements; i++) regularelements[i] = NULL;
		
		num_regularelements = 0;
	}

	RegularElementList(const RegularElementList &list) {
		num_regularelements = list.num_regularelements;
		max_regularelements = list.max_regularelements;
		regularelements = new RegularElement *[max_regularelements];

		for(int i = 0; i < num_regularelements; i++) {
			regularelements[i] = new RegularElement(*(list.regularelements[i]));
		}

		for(int i = num_regularelements; i < max_regularelements; i++) {
			regularelements[i] = NULL;
		}
	}

	void append(RegularElement *f) {
		if(num_regularelements >= max_regularelements) {
			max_regularelements += 20;
			RegularElement **flist = new RegularElement *[max_regularelements];
		
			for(int i = 0; i < num_regularelements; i++) flist[i] = regularelements[i];
			for(int i = num_regularelements; i < max_regularelements; i++) flist[i] = NULL;

			delete(regularelements);
			regularelements = flist;
		}

		/* add new RegularElement to list */

		regularelements[num_regularelements] = f;
		num_regularelements++;
	}

	void finalize() {
		if(this != NULL) {
			for(int i = 0; i < num_regularelements; i++) {
				if(regularelements[i]) delete(regularelements[i]);
			}

			delete(regularelements);
			delete(this);
		}
	}
};

class LineSegment {
public:
	int index;
	Triangle *tri;  // which triangle
	icVector2 start_coor, end_coor;  // all in local coordinate systems
	icVector2 start_pos, end_pos;
	double len;

public:
	LineSegment() {
		tri = NULL;
	}

	void calc_display_attributes() {
		tri->convert_local_to_global(start_coor, &start_pos);
		tri->convert_local_to_global(end_coor, &end_pos);
	}
};


class LineSegmentList {  // a separatrix
public:
	Singularity *start_s, *end_s;
	unsigned char direction; // 0 = follow the folow, 1 = reverse the flow
	double len;

	int num_linesegments;
	int max_linesegments;
	LineSegment **linesegments;

	LineSegmentList (int max) {
		max_linesegments = max;
		linesegments = new LineSegment *[max_linesegments];

		for(int i = 0; i < max_linesegments; i++) linesegments[i] = NULL;

		num_linesegments = 0;
	}

	void append(LineSegment *l) {
		if(num_linesegments >= max_linesegments) {
			max_linesegments += 200;
			LineSegment **llist = new LineSegment *[max_linesegments];

			for(int i = 0; i < num_linesegments; i++) {
				llist[i] = linesegments[i];
			}

			for(int i = num_linesegments; i < max_linesegments; i++) {
				llist[i] = NULL;
			}

			//delete(linesegments);
			delete[] linesegments;
			linesegments = llist;
		}

		linesegments[num_linesegments] = l;
		num_linesegments++;
	}

	void finalize() {
		if(this != NULL) {
			for(int i = 0; i < num_linesegments; i++) {
				if(linesegments[i]) delete(linesegments[i]);
			}

			delete [] linesegments;
			delete(this);
		}
	}
};


class SeparatricesList {
public:
	int num_separatrices;
	int max_separatrices;
	LineSegmentList **separatrices;

	SeparatricesList (int max) {
		max_separatrices = max;
		separatrices = new LineSegmentList *[max_separatrices];

		for(int i = 0; i < max_separatrices; i++) separatrices[i] = NULL;

		num_separatrices = 0;
	}

	void append(LineSegmentList *ls) {
		if(num_separatrices >= max_separatrices) {
			max_separatrices += 20;
			LineSegmentList **lslist = new LineSegmentList *[max_separatrices];
			
			for(int i = 0; i < num_separatrices; i++) lslist[i] = separatrices[i];
			for(int i = num_separatrices; i < max_separatrices; i++) lslist[i] = NULL;

			delete(separatrices);
			separatrices = lslist;
		}

		separatrices[num_separatrices] = ls;
		num_separatrices++;
	}

	void finalize() {
		if(this != NULL) {
			for(int i = 0; i < num_separatrices; i++) {
				if(separatrices[i]) separatrices[i]->finalize();
			}

			delete(separatrices);
			delete(this);
		}
	}
};


class Singularity {
public:
	int index;
	char vecten_index;
	unsigned char num_tensor_dir, vec_ten_mode, temp_use, singularity_type;  // 0 = source, 1=sink, 2=center, 3=source+center, 4=sink+center, 5=saddle
	icComplex eig_value[2];
	icVector2 pos, eig_dir[2];
	icVector2 tensor_dir[3];
	icVector2 center, dir[2], side_pt[4];
	icVector2 tensor_side_pt[3];
	Triangle *tri;
	SeparatricesList *separatrices_major, *separatrices_minor;


	Singularity() {
		separatrices_major = new SeparatricesList(2);
		separatrices_minor = new SeparatricesList(2);
		tri = NULL;
	}

	void calc_display_attributes() {
		tri->convert_local_to_global(pos, &center);
		center.entry[2] = -0.1;

		if(vec_ten_mode == 0) {
			dir[0] = tri->x_dir * eig_dir[0].entry[0] + tri->y_dir * eig_dir[0].entry[1];
			dir[1] = tri->x_dir * eig_dir[1].entry[0] + tri->y_dir * eig_dir[1].entry[1];
			side_pt[0] = center - mag_ratio * dir[0];
			side_pt[1] = center + mag_ratio * dir[0];
			side_pt[2] = center - mag_ratio * dir[1];
			side_pt[3] = center + mag_ratio * dir[1];
		}
		else {
			icVector2 temp_vec;
			
			for(int i = 0; i < num_tensor_dir; i++) {
				tri->convert_local_to_global_vector(tensor_dir[i], &temp_vec);
				tensor_side_pt[i] = center + mag_ratio * temp_vec;
			}
		}
	}

	void finalize() {
		separatrices_major->finalize();
		separatrices_minor->finalize();
		delete(this);
	}
};


class SingularityList {
public:
	int num_singularities;
	int max_singularities;
	Singularity **singularities;

	SingularityList (int max) {
		max_singularities = max;
		singularities = new Singularity *[max_singularities];

		for(int i = 0; i < max_singularities; i++) singularities[i] = NULL;
		num_singularities = 0;
	}

	void append(Singularity *s) {
		if(num_singularities >= max_singularities) {
			max_singularities += 20;
			Singularity **slist = new Singularity *[max_singularities];

			for(int i = 0; i < num_singularities; i++) slist[i] = singularities[i];
			for(int i = num_singularities; i < max_singularities; i++) slist[i] = NULL;

			delete(singularities);
			singularities = slist;
		}

		/* add new FixedPoint to list */
		singularities[num_singularities] = s;
		num_singularities++;
	}

	void insert_unique(Singularity *s){
		for(int i = 0; i < num_singularities; i++) {
			if(singularities[i]->tri == s->tri) return;
		}

		append(s);
	}

	void finalize() {
		if(this != NULL) {
			for(int i = 0; i<num_singularities; i++) {
				if(singularities[i]) singularities[i]->finalize();
			}

			//delete(singularities);
			delete[] singularities;
			delete(this);
		}
	}
};




struct SEInfo {
	list<SingularElement> fwd_elems;
	list<SingularElement> bwd_elems;
	list<SingularElement> kf;

	SEInfo() { clear(); }
	size_t size() const { return kf.size() + fwd_elems.size() + bwd_elems.size(); }

	void clear() {
		fwd_elems.clear();
		bwd_elems.clear();
		kf.clear();
	}
};


struct REInfo {
	list<RegularElement> fwd_elems;
	list<RegularElement> bwd_elems;
	list<RegularElement> kf;

	REInfo() { clear(); }
	size_t size() const { return kf.size() + fwd_elems.size() + bwd_elems.size(); }

	void clear() {
		fwd_elems.clear();
		bwd_elems.clear();
		kf.clear();
	}
};


struct DEInfo {
	SEInfo se;
	REInfo re;
	float weight_forward;
	bool diffusion;
	int region_id;

	DEInfo() { clear(); }
	void clear() { se.clear(); re.clear(); weight_forward = 0; diffusion = false; region_id = -2; }
	size_t size() const { return se.size() + re.size(); }
	size_t numKF() const { return se.kf.size() + re.kf.size(); }
	size_t numFwd() const { return se.fwd_elems.size() + re.fwd_elems.size(); }
	size_t numBwd() const { return se.bwd_elems.size() + re.bwd_elems.size(); }
};




class Polyhedron {
public:

//////////////////////mesh structures  /////////////////////////////////////////

	int index;
	unsigned char orientation;  // 0=ccw, 1=cw

	Vertex **vlist;    /* list of vertices */
	int nverts, max_verts;

	Triangle **tlist;  /* list of triangles */
	int ntris, max_tris;

	Edge **elist;      /* list of edges */
	int nedges, max_edges;

	Corner **clist;
	int ncorners, max_corners;

	VertexList **blist;
	int nboundaries, max_boundaries;

	icVector3 center;
	double radius, area;

	// initialization and finalization
	void initialize();
	void finalize();

	// I/O
	Polyhedron();
	Polyhedron(FILE *);
	void write_file(FILE *);

	// set up connectivity
	void create_edge(Vertex *, Vertex *);
	void create_edges();
	int face_to_vertex_ref(Triangle *, Vertex *);
	void order_vertex_to_tri_ptrs(Vertex *);
	void vertex_to_tri_ptrs();
	Triangle *find_common_edge(Triangle *, Vertex *, Vertex *);
	Triangle *other_triangle(Edge *, Triangle *);
	void create_pointers();
	void find_opposite_corners();
	void find_all_edges_around_vertex(int vertid);
	void determine_boundary_vertices();

	// compute geometry
	void average_normals();
	void calc_bounding_sphere();
	void calc_face_normals_and_area();
	void calc_edge_length();
	void calc_corner_angles();
	void set_local_frame();
	void calc_all_lamdas();

	// some tools
	double calc_triangle_area(icVector2 q0, icVector2 q1, icVector2 q2);
	void identify_corner_by_vertices(Vertex *v1, Vertex *v2, Corner **c);

////////////////////////// fields in general /////////////////////////////////////

	SingularityList *singularities;
	LineSegmentList *ls;
	unsigned int vec_ten_mode, original_vec_ten_mode;  // 0 = vector, 1 = tensor

	SingularElementList *singularelements;
	RegularElementList *regularelements;


	// initialization
	void reset_zero_field();
	//void removeElementsFromField();
	void copySingularElementList(SingularElementList **list);
	void copyRegularElementList(RegularElementList **list);
	void setSingularElementList(SingularElementList *list);
	void setRegularElementList(RegularElementList *list);

	// analysis
	void determine_singularities();
	void analyze_field(int do_analysis, int saddle_tri_id);

	// editing
	void global_rotate_field_tool(double global_rotate);
	void global_rotate_field(double global_rotate);
	void global_reflect_field_tool();
	void global_reflect_field();

	// smoothing flow inside a region
	VertexList *region_verts;
	CornerList *region_boundary;
  SkipList *vertex_queue;
	void construct_boundary_from_verts();
	void generate_region_from_boundary(TriangleList **tl);
	void search_path_within_on_path(Vertex *the_v, VertexList *the_vl, VertexList **result);
	void smooth_fld_in_region(int vec_ten_mode, TriangleList *tl);	
	void smooth_field();

	// common routines for pair cancellation and moving singularity
	void find_region_based_on_triangle_list(int direction, TriangleList *tl_in, TriangleList **tl_out);
	int contain_other_critical_pts(SingularityList *sl, TriangleList *tl, TriangleList *tl_constraint);
	int contain_only_required_singularities(SingularityList *sl, TriangleList *tl);
	void find_intersection(TriangleList *tl_in1, TriangleList *tl_in2, TriangleList **tl_out);
	void find_union(TriangleList *tl_in1, TriangleList *tl_in2, TriangleList **tl_out);
	void find_next_corner_for_edge(Corner *start_c, Corner **result_c);
	void find_corners_for_simply_connected_region(TriangleList *tl, CornerList **cl);
	int is_simply_connected(TriangleList *tl);
	int check_conley_condition(int which_operation, SingularityList *sl, TriangleList **tl);
	int find_region_for_attractor(Singularity *source, Triangle *the_tri, LineSegmentList *vertical[2], TriangleList *tl_constraint, TriangleList **tl_attractor);
	void find_region_for_repeller(Singularity *source, Triangle *the_tri, TriangleList **region_repellor);

	// singularity pair cancellation
	void find_base_region_based_on_sepatatrix(LineSegmentList *sepa, double percentage, TriangleList **tl_base);
	int find_common_region_for_pair_cancellation(Singularity *index_neg, Singularity *index_pos, TriangleList **tl);
	int singularity_pair_cancellation(int cp1, int cp2, double rotate_angle);
	void singularity_pair_cancellation_field(int cp1, int cp2);

	// moving singularity
	LineSegmentList *move_trace[3];
	int build_traces_for_moving(Triangle *the_tri, icVector2 pos, Singularity *source, double percentage);
	int find_common_region_for_moving_source(Singularity *index_pos, Triangle *the_tri, TriangleList **tl);
	int singularity_movement(int sid, int tid, double rotate_angle);
	void singularity_movement_field(int cp1, int tid);

	void undo();
	void redo();

	// visualization
	//bool ignore_at_singularity;
	void calc_new_texture_coors();

	// I/O
	void get_model_from_file(int contain_vector_data);
	void write_fld_ctrl_file(char name[128]);
	void read_fld_ctrl_file(char name[128]);

	// utilities
	void convert_vecfld_to_tenfld_tool();
	void convert_vecfld_to_tenfld();
	void convert_tenfld_to_vecfld_tool();
	void convert_tenfld_to_vecfld();

////////////////////////// vector field  /////////////////////////////////////////

	// initialization
	void update_vecfld(int target_region_id, int do_set_zero, const std::vector<const DEInfo*> &design_elements);

	// analysis
	void handle_transversity();
	int trace_trajectory_in_triangle_vecfld(icVector2 start_pos, unsigned char direction, Triangle *tri, Edge **incoming_edge, Triangle **next_tri, icVector2 *next_start_pos, LineSegmentList **ls);
	int trace_trajectory_in_domain_vecfld(int max_triangles, icVector2 start_pos, unsigned char direction, Triangle *tri, LineSegmentList **ls);
	void compute_vector_separatrices(int saddle_tri_id);


////////////////////////// tensor field  /////////////////////////////////////////

	// initialization
	void update_tenfld(int target_region_id, int do_set_zero, double weight_interpolate, const vector<const DEInfo*> &design_elements);

	// analysis
	int trace_trajectory_in_triangle_tenfld(icVector2 start_pos, unsigned char direction, icVector2 initial_dir, Triangle *tri, Edge **incoming_edge, Triangle **next_tri, icVector2 *next_start_pos, icVector2 *next_initial_dir, LineSegmentList **ls);
	int trace_trajectory_in_domain_tenfld(int max_triangles, icVector2 start_pos, unsigned char direction, icVector2 initial_dir, Triangle *tri, LineSegmentList **ls);
	void compute_tensor_separatrices();


// =========== vector field input/output (beginning) =========================================
	void get_model_from_file();
	void write_vecfld_ctrl_file(const char name[128]);
	void read_vecfld_ctrl_file(const char name[128]);

	void write_tenfld_ctrl_file(const char name[128]);
	void read_tenfld_ctrl_file(const char name[128]);


	int trace_one_trajectory_in_one_triangle(icVector2 start_pos, unsigned char direction, Triangle *tri, Edge **incoming_edge, Triangle **next_tri, icVector2 *next_start_pos, LineSegmentList **ls);
	int trace_one_tensor_trajectory_in_one_triangle(icVector2 start_pos, unsigned char direction, icVector2 initial_dir, Triangle *tri, Edge **incoming_edge, Triangle **next_tri, icVector2 *next_start_pos, icVector2 *next_initial_dir, LineSegmentList **ls);
// if return 0 hit something, if return 1, hit boundary of the domain
	int trace_one_trajectory_in_domain(int num_triangles, icVector2 start_pos, unsigned char direction, Triangle *tri, LineSegmentList **ls);
	int trace_one_tensor_trajectory_in_domain(int num_triangles, icVector2 start_pos, unsigned char direction, icVector2 initial_dir, Triangle *tri, LineSegmentList **ls);

// =========== vector field analysis (end)       =========================================

	void retrieve_divergence_field();
	int trace_trajectory_in_triangle_vecfld(icVector2 start_pos, unsigned char direction, Triangle *tri, Edge **incoming_edge, Triangle **next_tri, icVector2 *next_start_pos, icVector2List **vlist);
	int trace_trajectory_in_triangle_tenfld(icVector2 start_pos, unsigned char direction, icVector2 initial_dir, Triangle *tri, Edge **incoming_edge, Triangle **next_tri, icVector2 *next_start_pos, icVector2 *next_initial_dir, icVector2List **vlist);


	// apply design elements to current matrix value
	icMatrix2x2 apply_T(const Vertex &v, const list<SingularElement> &se, const list<RegularElement> &re, float weight_interpolate);


	PlyOtherProp *vert_other,*face_other;
	int regionID;
};


#endif /* __TENDES_PLANE_H__ */

